"""
Fixed version of `torchrun` on Jülich Supercomputing Centre for PyTorch
versions <2. Requires Slurm usage.

To use, modify your execution like the following:

Old
```shell
torchrun [...]
# or
python -m torch.distributed.run [...]
```

New
```shell
python /path/to/torchrun_jsc/run_old.py [...]
# or if `torchrun_jsc` is on `PYTHONPATH`
python -m torchrun_jsc.run_old [...]
```

Tested for PyTorch <2, 2.1.2
"""

from argparse import ArgumentParser
import ipaddress
import runpy
import socket

from torch.distributed.elastic.agent.server import api as sapi

from . import arg_patching
from . import parsing


def parse_args():
    parser = ArgumentParser()
    parser.add_argument('--rdzv_endpoint', '--rdzv-endpoint')
    parser.add_argument('--rdzv_conf', '--rdzv-conf')
    args = parser.parse_known_args()[0]

    endpoint = args.rdzv_endpoint
    host = parsing.parse_host(endpoint)

    conf = args.rdzv_conf
    is_host = parsing.parse_is_host(conf)

    return host, conf, is_host


def fix_torch_run(host):
    _orig_get_fq_hostname = sapi._get_fq_hostname

    if host:
        try:
            ipaddress.ip_address(host)
            is_ip = True
        except ValueError:
            is_ip = False

        if is_ip:
            def new_get_fq_hostname():
                return socket.gethostbyaddr(host)[0]
        else:
            def new_get_fq_hostname():
                return socket.getfqdn(host)
    else:
        new_get_fq_hostname = _orig_get_fq_hostname

    sapi._get_fq_hostname = new_get_fq_hostname


def main():
    host, conf, is_host = parse_args()
    arg_patching.fix_is_host(is_host, conf)
    fix_torch_run(host)
    runpy.run_module('torch.distributed.run', run_name='__main__')


if __name__ == '__main__':
    main()



"""
Common parsing routines for `torchrun` arguments.
"""


def _as_bool(key, value):
    if value is None or isinstance(value, bool):
        return value
    if isinstance(value, int):
        if value == 1:
            return True
        if value == 0:
            return False
    elif isinstance(value, str):
        if value.lower() in ['1', 'true', 't', 'yes', 'y']:
            return True
        if value.lower() in ['0', 'false', 'f', 'no', 'n']:
            return False
    raise ValueError(
        f'The rendezvous configuration option {key} does not represent a '
        f'valid boolean value.'
    )


def parse_host(rdzv_endpoint):
    host = (
        rdzv_endpoint.rsplit(':', 1)[0]
        if rdzv_endpoint
        else None
    )
    return host


def parse_is_host(rdzv_conf):
    is_host = None
    if rdzv_conf is not None:
        confs = rdzv_conf.split(',')
        for (key, value) in map(lambda kv: kv.split('=', 1), confs):
            if key == 'is_host':
                is_host = _as_bool(key, value)
                break
    return is_host



"""
Fixed version of `torchrun` on Jülich Supercomputing Centre for PyTorch
versions ≥1.9. Requires Slurm usage.

To use, modify your execution like the following:

Old
```shell
torchrun [...]
# or
python -m torch.distributed.run [...]
```

New
```shell
# if `torchrun_jsc` is on `PYTHONPATH`
python -m torchrun_jsc [...]
# or if `torchrun_jsc` is `pip`-installed
torchrun_jsc [...]
```
"""

from packaging import version

import torch


def main():
    torch_ver = version.parse(torch.__version__)
    if torch_ver.major >= 2:
        if torch_ver.major > 2:
            print(
                'WARNING: This version of PyTorch is not officially supported '
                'by `torchrun_jsc`. You may be able to ignore this warning.'
            )

        from .run_old import main as run_main_old
        run_main_old()
    elif torch_ver.major == 1 and torch_ver.minor >= 9:
        from .run_old import main as run_main_old
        run_main_old()
    else:
        print(
            'This version of PyTorch is not supported by `torchrun_jsc` '
            'because it does not have the `torchrun` API implemented. '
            'Please use another launch API.'
        )
        exit(1)



"""
Common patching routines for `torchrun` arguments.
"""

import os
import sys


def fix_is_host(is_host, conf):
    if is_host is None:
        slurm_is_host = int(os.getenv('SLURM_NODEID', '0') == '0')

        if not conf:
            insertion_index = min(len(sys.argv), 1)
            sys.argv.insert(insertion_index, '--rdzv_conf=')

        # Since `torchrun` only uses standard `argparse` for
        # parsing, we do not need to worry about discerning multiple
        # `--rdzv_conf` arguments (one for `torchrun`, one for the
        # script).
        for (i, arg) in enumerate(sys.argv):
            if (
                    arg.startswith('--rdzv_conf')
                    or arg.startswith('--rdzv-conf')
            ):
                # Handle specification as two arguments vs. as one
                # argument.
                if arg in ['--rdzv_conf', '--rdzv-conf']:
                    modification_index = i + 1
                    old_conf = sys.argv[modification_index]
                else:
                    modification_index = i
                    old_conf = (
                        sys.argv[modification_index].split('=', 1)[1])

                # Handle empty conf specification.
                if old_conf:
                    sys.argv[modification_index] = (
                        f'{sys.argv[modification_index]},')
                sys.argv[modification_index] = (
                    f'{sys.argv[modification_index]}'
                    f'is_host={slurm_is_host}'
                )
                break



from setuptools import setup

setup(
    name='torchrun_jsc',
    python_requires='>=3.6,<4',
    version='0.0.3',
    install_requires=[
        'packaging',
        # PyTorch 1.9 introduced the `torchrun` API.
        #
        # Since that version is already quite old in the fast-moving
        # field of deep learning and this package is concerned only with
        # fixing `torchrun` issues, we do not support earlier versions
        # with the old `torch.distributed.launch` API.
        #
        # We allow any future PyTorch versions but print a warning if a
        # not officially supported version is used.
        'torch>=1.9',
    ],
    entry_points={
        "console_scripts": [
            "torchrun_jsc = torchrun_jsc.run:main",
        ],
    },
    author='Jan Ebert',
    author_email='ja.ebert@fz-juelich.de',
    description=(
        'Fixed version of `torchrun` on Jülich Supercomputing Centre. '
        'Requires Slurm usage.'
    ),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    url='https://github.com/HelmholtzAI-FZJ/torchrun_jsc',
)



"""
WARNING: Currently broken, do not use. Leads to indeterministic
deadlocks.

Fixed version of `torchrun` on Jülich Supercomputing Centre for PyTorch
versions ≥2. Requires Slurm usage.

To use, modify your execution like the following:

Old
```shell
torchrun [...]
# or
python -m torch.distributed.run [...]
```

New
```shell
python /path/to/torchrun_jsc/run.py [...]
# or if `torchrun_jsc` is on `PYTHONPATH`
python -m torchrun_jsc.run [...]
```
"""

from argparse import ArgumentParser
import runpy
import sys

from . import arg_patching
from . import parsing


def parse_args():
    parser = ArgumentParser()
    parser.add_argument('--rdzv_endpoint', '--rdzv-endpoint')
    parser.add_argument('--rdzv_conf', '--rdzv-conf')
    parser.add_argument('--local_addr', '--local-addr')
    args = parser.parse_known_args()[0]

    endpoint = args.rdzv_endpoint
    host = parsing.parse_host(endpoint)

    conf = args.rdzv_conf
    is_host = parsing.parse_is_host(conf)

    local_addr = args.local_addr

    return host, conf, is_host, local_addr


def fix_get_hostname(host, local_addr):
    if host and not local_addr:
        insertion_index = min(len(sys.argv), 1)
        sys.argv.insert(insertion_index, f'--local_addr={host}')


def main():
    host, conf, is_host, local_addr = parse_args()
    fix_get_hostname(host, local_addr)
    arg_patching.fix_is_host(is_host, conf)
    runpy.run_module('torch.distributed.run', run_name='__main__')


if __name__ == '__main__':
    main()
